# assaydata.py - script functions developed for integration with ScienceCloud via PipelinePilot server
# Use of this code is subject to the terms of the StarDrop License Agreement and use, copying or redistribution of this file for use by other persons or organisations is prohibited
# Copyright(C) Optibrium Ltd 2017
# stardrop-support@optibrium.com
#
"""
Module handles information about assays accessed via ScienceCloud.
"""

import PythonQt
import json
import re


class AssayData:
    """Class to hold information about how to process results for a particular assay."""
    def __init__(self, name):
        """Create an AssayData object.
        @param: name: A long name (should be unique) constructed from a number of fields in the result data.
        """
        self._name = name
        self._units = name.split('/')[-3]#The units entry
        self._alias = name
        self._factor = self._units in ('uM','nM','mM','pM')
        self._uncertainty = None
        self._add_count = False
        self._add_summary = False
        
    def __hash__(self):
        """Override hash to allow use of this class as an element of a set.
        @return: A hash derived from the long name.
        """
        return hash(self._name)
        
    def __eq__(self,other):
        """Override eq to allow use of this class as an element of a set.
        @return: True if hashes match..
        """
        return self.__hash__() == other.__hash__()
        
    def name(self):
        """Return the name used for display.
        @return: A text string containing the display name.
        """
        return self._alias
        
    def units(self):
        """Return units defined for this assay.
        @return: Astring representation of the units.
        """
        return self._units
        
    def defaultUncertainty(self):
        """Return a default value to use for uncertainty in this assay.
        @return: A float value.
        """
        if self._uncertainty:
            return self._uncertainty
        if self._factor:
            return 1.0
        return 0.0
        
    def showCount(self):
        """Return whether to display a column containing a count of values.
        @return: True if count should be displayed.
        """
        return self._add_count
        
    def showSummary(self):
        """Return whether to display a column containing a summary of values.
        @return: True if summary should be displayed.
        """
        return self._add_summary
        
    def isFactor(self):
        """Check whether the assay has a factor uncertainty (ie uses a geometric mean).
        @return: True if assay uses a factor.
        """
        return self._factor
        
    def assayID(self):
        """Return the assay's unique iedntifier.
        @return: A string containing the id.
        """
        return self._name
        
    def toJson(self, selections):
        """Generate a JSON-compatible dictionary representing the assay.
        @return: A dictionary containing assay properties.
        """
        json_assay = {}
        json_assay["alias"] = self._alias
        json_assay["default_uncertainty"] = self.defaultUncertainty()
        json_assay["include_count"] = self._add_count
        json_assay["include_values"] = self._add_summary
        json_assay["factor"] = self._factor
        json_assay["selected"] = len(selections) == 0 or self._name in selections
        return json_assay
        
    def update(self,json_assay):
        """Update the assay details from a JSON object.
        @param: json_assay: A JSON representation of the edited asay details.
        """
        if "alias" in json_assay:
            self._alias = json_assay["alias"]
        if "default_uncertainty" in json_assay:
            self._uncertainty = json_assay["default_uncertainty"]
        if "include_count" in json_assay:
            self._add_count = json_assay["include_count"]
        if "include_values" in json_assay:
            self._add_summary = json_assay["include_values"]
        if "factor" in json_assay:
            self._factor = json_assay["factor"]
        
class AssayManager:
    """Class to handle management of AssayData objects."""
    def __init__(self):
        """Initialise an AssayManager."""
        self._assays = {}
        
    def __iter__(self):
        """Override iter to allow itetration of the stored AssayData objects."""
        return iter(self._assays)
        
    def __len__(self):
        """Override len to allow itetration of the stored AssayData objects."""
        return len(self._assays)
        
    def __getitem__(self,key):
        """Override getitem to allow itetration of the stored AssayData objects."""
        return self._assays[key]
        
    def setAssayData(self,key,assay):
        """Specify the data for an assay. Overwrites any existing entry.
        @param: key: The unique identifier for the asssy.
        @param: assay: An AssayData object.
        """
        self._assays[key] = assay
        
    def loadFromFile(self):
        """Load the previously defined assay data from a file."""
        try:
            with open('sciencecloudassays.txt') as input_file:
                assays = json.loads(input_file.read())
                for assay in assays:
                    assay_data = AssayData(assay)
                    assay_data.update(assays[assay])
                    self._assays[assay] = assay_data
        except:
            pass
    def serialise(self):
        """Write the current assay data to a file."""
        try:
            with open('sciencecloudassays.txt','w') as output_file:
                assays = {}
                for assay in self._assays:
                    assays[assay] = self._assays[assay].toJson([])
                output_file.write(json.dumps(assays))
        except Exception as e:
            print(e)
            pass
                    
assay_manager = AssayManager()
assay_manager.loadFromFile()
        
        
def createAssayData(entries,assay_indexes):
    """Create an AssayData object from ScienceCloud data, or retrieve from the existing store.
    @return: An AssayData object with an appropriate key value.
    """
    assay = AssayData('/'.join([entries[i] for i in assay_indexes]))
    if assay in assay_manager:
        return assay_manager[assay]
    return assay
    
def jsoniseAssays(assays,selections):
    """Convert a list of assays to JSON format for passing to the UI.
    @param: assays: A list of AssayData objects.
    @param: selections: A list of assay identifiers that should be selected when shown to the user (used in refresh, otherwise empty).
    @return: A dictionary suitable for conversion to JSON.
    """
    json_assays = {}
    for assay in assays:
        # ignore 'no data' assay details - '//////////'
        if not bool(re.match('^[/]+$', assay.assayID())):
            json_assays[assay.assayID()] = assay.toJson(selections)
    return json_assays
    
def selectAssays(stardrop,assays,selections,silent):
    """Allow user to select and configure the assays to include in a data set.
    @param: stardrop: The StarDrop application interface.
    @param: A list of AssayData objects representing the available assays for this download.
    @param: selections: A list of assay identifiers that should be selected when shown to the user (used in refresh, otherwise empty).
    @param: silent: Flag indicating if user should be shown the assay selection dialog again when refreshing
    @return: A tuple containing a flag and a set of AssayData objects representing the assays to be included in the dataset.
    """
    selected_assays = set()
    assay_json = jsoniseAssays(assays,set(selections))
    if len(assay_json) > 0:
        dlg = PythonQt.ScienceCloudDisplay(stardrop.mainWindow())
        dlg.setDisplayData(json.dumps(assay_json))
        if not silent:
            if dlg.run():
                assay_json = json.loads(dlg.displayData().data())
            else:
                return False,selected_assays
        for assay in assay_json:
            if assay_json[assay]["selected"]:
                new_data = AssayData(assay)
                new_data.update(assay_json[assay])
                selected_assays.add(new_data)
                #update cache as well
                assay_manager.setAssayData(assay,new_data)
        assay_manager.serialise()
  
        dlg.delete()
    return True,selected_assays
    
