# datasetcreator.py - script functions developed for integration with ScienceCloud via PipelinePilot server
# Use of this code is subject to the terms of the StarDrop License Agreement and use, copying or redistribution of this file for use by other persons or organisations is prohibited
# Copyright(C) Optibrium Ltd 2017
# stardrop-support@optibrium.com
#

import app
import math

       
def _calculateStats(values,default_error,is_factor):
    """Calculate an appropriate mean and stabdard deviation from the experimental values.
    @param: values: A list of floats representing the values (these will have already been logged if we are working with a geometric mean).
    @param: default_error: The standard deviation to use if we have less than 3 data points.
    @param: is_factor: Flag to indicate oif the error is a factor (ie we are working with geometric data)
    @return: Tuple containing a value and an error.
    """
    val = 0.0
    count = len(values)
    error = default_error
    sum = 0.0
    for v in values:
        sum += v
    val = sum/count
    if count > 2:
        sumsqr = 0.0
        for v in values:
            sumsqr += pow(v-val,2)
        error = math.sqrt(sumsqr/count)
        if is_factor:
            error = math.exp(error)
    if is_factor:
        val = math.exp(val)             
    return val,error  
    
def _getNumericalEntries(data, assay):
    """Extract a suitable set of entries from the data points for a given assay.
    @param: data: TA list of experimental results for tis compound and this sssay.
    @param: assay: An AssayData object describing how to process the data.
    @return: A list of StarDrop data entries, one for the assay result and optionally one containing a count and one containing a text representation of the individual results.
    """
    def extractValue(value,is_factor):
        val = float(value.lstrip("<=>"))
        if is_factor:
            val = math.log(val)
        return val
    def massageQualifier(data_entry):
        """Fix qualifier problem that affects functions"""
        if data_entry[2][0] == '0':
            value = ('\x00',data_entry[2][1],data_entry[2][2])
            return (data_entry[0],data_entry[1],value)
        return data_entry
    count = 0
    text_values = []
    numerical_values = []
    entries = []
    for result in data:
        try:
            value = extractValue(result["Value"],assay.isFactor())
            count += 1
            numerical_values.append(value)
            text_values.append(result["Operator"] + result["Value"])
        except:
            pass
    if count == 0:
        entries.append(app.Dataset.createInvalidEntry(app.cont))
    else:
        val,error = _calculateStats(numerical_values,assay.defaultUncertainty(), assay.isFactor())
        entries.append(massageQualifier(app.Dataset.cont(val,error)))
    if assay.showCount():
        entries.append(massageQualifier(app.Dataset.cont(float(count),0.0)))
    if assay.showSummary():
        entries.append(app.Dataset.string(" ".join([tv.lstrip('=') for tv in text_values])))
    return entries
            
def _unitsLookup(units_as_text):
    """Convert a text representation of a unit to a StarDrop integer.
    @param: units_as_text: The value from ScienceCloud:
    @param: An integer to use when constructing a StarDrop column.
    @TODO: add additional units here
    """
    units = 0
    if units_as_text == "uM":
        units = 1
    elif units_as_text == "nM":
        units = 3
    elif units_as_text == "%":
        units = 8
    elif units_as_text == "mM":
        units = 18
    elif units_as_text == "pM":
        units = 20
    return units
    
def createDataset(assays,data,aggregation_mode,include_libraries = True):
    """Create a StarDrop dataset from a list of required assays and the data from ScienceCloud.
    @param: assays: A list of AssayData objects representing columns to include in the data set.
    @param: data: The data for each row as a dictionary of property to values. For assays, these values need to be aggregated.
    @param: aggregation_mode: The type of aggregation to use (compound or batch).
    @param: include_libraries: Flag indicating whether result should include a column showing the library
    @return: A StarDrop dataset.
    """
    cols = []
    molcol = app.Dataset.meta("Structure",app.mol)
    primary_id_name = "Compound ID"
    secondary_id_name = "Batch IDs"
    if aggregation_mode == "batch":
        primary_id_name = "Batch ID"
        secondary_id_name = "Compound ID"
    idcol = app.Dataset.meta(primary_id_name)
    second_idcol = app.Dataset.meta(secondary_id_name)
    compound_tag_col = app.Dataset.meta("Compound Tag")
    batch_tag_col = app.Dataset.meta("Batch Tag")
    cols.append(molcol)
    cols.append(idcol)
    cols.append(second_idcol)
    cols.append(compound_tag_col)
    cols.append(batch_tag_col)
    
    for assay in assays:
        units = _unitsLookup(assay.units())
        cols.append(app.Dataset.createNumberHeader(assay.name(),units = units, factor = assay.isFactor()))
        if assay.showCount():
            cols.append(app.Dataset.createNumberHeader(assay.name() + "_Count"))
        if assay.showSummary():
            cols.append(app.Dataset.meta(assay.name() + "_Values"))
    if include_libraries:
        cols.append(app.Dataset.meta("Library"))
        
    dataset = app.Dataset(cols)
    dataset.idColName = primary_id_name
    for id in data:
        #check that there is an assay at 0, not secondary_ids
        #assay might be blank here for molecules with no assay data: in such cases the molecule is 
        #still created, but there will be no assay column created below as 'assays' is the empty set.
        for assay in data[id].keys():
            try:
                dataset.cols[0].append(app.Dataset.molecule(data[id][assay][0]["Smiles"]))
                break
            except:
                continue
        dataset.cols[1].append(app.Dataset.string(id))
        dataset.cols[2].append(app.Dataset.string(' '.join(data[id]["secondary_ids"])))
        dataset.cols[3].append(app.Dataset.string(data[id]["compound_tag"]))
        dataset.cols[4].append(app.Dataset.string(','.join([tag for tag in data[id]["batch_tag"] if tag != ""])))
        count = 5
        for i,assay in enumerate(assays):
            entries = _getNumericalEntries(data[id].get(assay,{}),assay)
            for entry in entries:
                dataset.cols[count].append(entry)
                count += 1
        if include_libraries:  
            dataset.cols[-1].append(app.Dataset.string(','.join(data[id]["libraries"]))) 
    return dataset
